@@ -7,7 +7,6 @@
 #
 
 import json
-import logging
 from datetime import date, datetime
 from functools import partial
 
@@ -355,11 +354,6 @@ class PerspectiveWidget(DOMWidget, PerspectiveViewer):
             # Viewer will ignore **options if `data` is a Table or View.
             super(PerspectiveWidget, self).load(data, **options)
 
-            # Do not enable editing if the table is unindexed.
-            if self.editable and self.table.get_index() is None:
-                logging.critical("Cannot edit on an unindexed `perspective.Table`!")
-                self.editable = False
-
         # Notify front-end of load immediately.
         message = self._make_load_message()
         self.send(message.to_dict())
